@@ -3,8 +3,29 @@
 ;;; (load "repl.scm") ((*repl* 'run))
 
 (provide 'repl.scm)
+
+(when (= (*s7* 'print-length) 12)    ; old default value
+  (set! (*s7* 'print-length) 32))
+
+(when (file-exists? ".repl")         ; local (scheme) initialization file for repl
+  (load ".repl"))
+
+(set! (*s7* 'history-enabled) #f)
+
+(unless (defined? 'stderr-buffered)
+  (define stderr-buffered #f))
+
+(define (fformat port . args)
+  (apply format port args)
+  (when (and stderr-buffered
+	     (eq? port *stderr*))
+    (flush-output-port *stderr*)))
+
 (require libc.scm)
 
+(define old-debug (*s7* 'debug))
+(set! (*s7* 'debug) 0)
+
 (unless (defined? '*repl*)
   (define *repl*                    ; environment that holds the REPL functions
     (let ((prompt #f)               ; function to get/set prompt
@@ -92,54 +113,55 @@
 		     (string=? (vector-ref histbuf i) line))
 		 (and (< i histsize) i))))
 	  
-	  (define history-size (dilambda
-				(lambda ()
-				  histsize)
-				(lambda (new-size)
-				  (unless (= new-size histsize)
-				    (if (<= new-size 0)
-					(error 'out-of-range "new history buffer size must be positive")
-					(let ((new-hist (make-vector new-size ""))
-					      (new-end (min (- new-size 1) histpos)))
-					  (let loop ((oldpos histpos)
-						     (newpos new-end))
-					    (set! (new-hist newpos) (histbuf oldpos))
-					    (set! newpos (- (if (zero? newpos) new-size newpos) 1))
-					    (unless (= newpos new-end)
-					      (set! oldpos (- (if (zero? oldpos) histsize oldpos) 1))
-					      (unless (= oldpos histpos)
-						(loop oldpos newpos))))
-					  (set! histsize new-size)
-					  (set! histpos new-end)
-					  (set! histbuf new-hist)
-					  new-size))))))
+	  (define history-size 
+	    (dilambda
+	     (lambda ()
+	       histsize)
+	     (lambda (new-size)
+	       (unless (= new-size histsize)
+		 (if (<= new-size 0)
+		     (error 'out-of-range "new history buffer size must be positive")
+		     (let ((new-hist (make-vector new-size ""))
+			   (new-end (min (- new-size 1) histpos)))
+		       (let loop ((oldpos histpos)
+				  (newpos new-end))
+			 (set! (new-hist newpos) (histbuf oldpos))
+			 (set! newpos (- (if (zero? newpos) new-size newpos) 1))
+			 (unless (= newpos new-end)
+			   (set! oldpos (- (if (zero? oldpos) histsize oldpos) 1))
+			   (unless (= oldpos histpos)
+			     (loop oldpos newpos))))
+		       (set! histsize new-size)
+		       (set! histpos new-end)
+		       (set! histbuf new-hist)
+		       new-size))))))
+	  
+	  (define (history back)
+	    (let ((i (+ histpos back)))
+	      (copy (histbuf (if (< i 0)
+				 (+ histsize i)
+				 (if (>= i histsize)
+				     (- i histsize)
+				     i))))))
 
-	  (define history (dilambda 
-			   (lambda (back)
-			     (let ((i (+ histpos back)))
-			       (copy (histbuf (if (< i 0)
-						  (+ histsize i)
-						  (if (>= i histsize)
-						      (- i histsize)
-						      i))))))
-			   (lambda (new-line)
-			     (let ((pos (history-member new-line)))
-			       (when (integer? pos)                   ; remove the earlier case, circularly compress the buffer
-				 (when (>= pos histpos)
-				   (do ((i pos (+ i 1)))
-				       ((>= i (- histsize 1)))
-				     (set! (histbuf i) (histbuf (+ i 1))))
-				   (set! (histbuf (- histsize 1)) (histbuf 0))
-				   (set! pos 0))
-				 (do ((i pos (+ i 1)))
-				     ((>= i (- histpos 1)))
-				   (set! (histbuf i) (histbuf (+ i 1))))
-				 (set! histpos (- histpos 1))))
-
-			     (set! (histbuf histpos) (copy new-line))
-			     (set! histpos (+ histpos 1))
-			     (if (= histpos histsize)
-				 (set! histpos 0)))))
+	  (define (set-history! new-line)
+	    (let ((pos (history-member new-line)))
+	      (when (integer? pos)                   ; remove the earlier case, circularly compress the buffer
+		(when (>= pos histpos)
+		  (do ((i pos (+ i 1)))
+		      ((>= i (- histsize 1)))
+		    (set! (histbuf i) (histbuf (+ i 1))))
+		  (set! (histbuf (- histsize 1)) (histbuf 0))
+		  (set! pos 0))
+		(do ((i pos (+ i 1)))
+		    ((>= i (- histpos 1)))
+		  (set! (histbuf i) (histbuf (+ i 1))))
+		(set! histpos (- histpos 1))))
+	    
+	    (set! (histbuf histpos) (copy new-line))
+	    (set! histpos (+ histpos 1))
+	    (if (= histpos histsize)
+		(set! histpos 0)))
 	  
 	  (define (history-help)
 	    (set! (*repl* 'helpers)
@@ -214,64 +236,16 @@
 			   (set! (h 'result) (((rootlet) 'system) cur-line #f))
 			   (set! (h 'result) #f)))))))
 	    
-	    (define (with-repl-let body)
-	      ;; for multiline edits, we will use *missing-close-paren-hook* rather than try to parse the input ourselves.
-	      (let ((old-badexpr-hook (hook-functions *missing-close-paren-hook*))
-		    (old-unbound-var-hook (hook-functions *unbound-variable-hook*))
-		    (old-eval #f)
-		    (old-eval-string #f)
-		    (old-load #f))
-		
-		;; if the repl's top-level-let is not rootlet, and we load some file into rootlet
-		;;   that (for example) defines a function at its top level, that function's definition
-		;;   env is rootlet.  If we then define something in the repl, it is in the
-		;;   repl's top level.  If we now call the function asking is the new thing defined,
-		;;   it looks in rootlet, not the repl top-level, and says no.
-		;; so, locally redefine these three to use the repl top-level while we're in the repl.
-		;; in addition, for each of these, we need to report missing close parens and so on
-		
-		(let ((repl-hooks
-		       (lambda ()
-			 (set! (hook-functions *missing-close-paren-hook*) (cons badexpr old-badexpr-hook))
-			 (set! (hook-functions *unbound-variable-hook*) (cons shell? old-unbound-var-hook))))
-		
-		      (original-hooks
-		       (lambda ()
-			 (set! unbound-case #f)
-			 (set! (hook-functions *missing-close-paren-hook*) old-badexpr-hook)
-			 (set! (hook-functions *unbound-variable-hook*) old-unbound-var-hook))))
-		
-		  (let ((new-load 
-			 (let ((+documentation+ "this is the repl's load replacement; its default is to use the repl's top-level-let.")
-			       (+signature+ '(values string? let?)))
-			   (lambda* (file (e (*repl* 'top-level-let)))
-			     (dynamic-wind repl-hooks (lambda () (load file e)) original-hooks))))
-			
-			(new-eval 
-			 (let ((+documentation+ "this is the repl's eval replacement; its default is to use the repl's top-level-let.")
-			       (+signature+ '(values list? let?)))
-			   (lambda* (form (e (*repl* 'top-level-let)))
-			     (dynamic-wind repl-hooks (lambda () (eval form e)) original-hooks))))
-			
-			(new-eval-string 
-			 (let ((+documentation+ "this is the repl's eval-string replacement; its default is to use the repl's top-level-let.")
-			       (+signature+ '(values string? let?)))
-			   (lambda* (str (e (*repl* 'top-level-let)))
-			     (dynamic-wind repl-hooks (lambda () (eval-string str e)) original-hooks)))))
-		    
-		    (dynamic-wind
-			(lambda ()
-			  (repl-hooks)
-			  (set! eval new-eval)
-			  (set! eval-string new-eval-string)
-			  (set! load new-load))
-			body
-			(lambda ()
-			  (set! eval old-eval)
-			  (set! eval-string old-eval-string)
-			  (set! load old-load)
-			  (original-hooks)))))))
-	    
+	    (define new-eval 
+	      (let ((+documentation+ "this is the repl's eval replacement; its default is to use the repl's top-level-let.")
+		    (+signature+ '(values #t let?)))
+		(lambda (form . rest) ; use lambda (not lambda*) so we can handle forms like :key
+		  (let ((e (if (pair? rest) 
+			       (car rest)
+			       (*repl* 'top-level-let))))
+		    (let-temporarily (((hook-functions *unbound-variable-hook*) (list shell?)) ; so pwd et al will work
+				      ((*s7* 'history-enabled) #t))
+		      (eval form e))))))
 	    
 	    ;; -------- match parens --------
 	    (define check-parens
@@ -382,28 +356,28 @@
 		      text #\escape))
 	    
 	    (define (move-cursor y x)
-	      (format *stderr* "~C[~D;~DH" #\escape y x))
+	      (fformat *stderr* "~C[~D;~DH" #\escape y x))
 
 	    (define (cursor-coords)
 	      (let* ((c (string #\null #\null))
 		     (cc (string->c-pointer c))
 		     (terminal-fd (fileno stdin)))
-		(format *stderr* "~C[6n" #\escape)
+		(fformat *stderr* "~C[6n" #\escape)
 		(do ((b (read terminal-fd cc 1) (read terminal-fd cc 1)))
-		    ((char=? (c 0) #\escape)))
+		    ((char=? (string-ref c 0) #\escape)))
 		(read terminal-fd cc 1) 
-		(and (char=? (c 0) #\[)
+		(and (char=? (string-ref c 0) #\[)
 		     (let ((y 0)
 			   (x 0))
 		       (do ((b (read terminal-fd cc 1) (read terminal-fd cc 1)))
-			   ((not (char-numeric? (c 0))))
-			 (set! y (- (+ (* 10 y) (char->integer (c 0))) (char->integer #\0))))
-		       (and (char=? (c 0) #\;)
+			   ((not (char-numeric? (string-ref c 0))))
+			 (set! y (- (+ (* 10 y) (char->integer (string-ref c 0))) (char->integer #\0))))
+		       (and (char=? (string-ref c 0) #\;)
 			    (do ((b (read terminal-fd cc 1) (read terminal-fd cc 1)))
-				((not (char-numeric? (c 0)))
-				 (and (char=? (c 0) #\R)
+				((not (char-numeric? (string-ref c 0)))
+				 (and (char=? (string-ref c 0) #\R)
 				      (cons x y)))
-			      (set! x (- (+ (* 10 x) (char->integer (c 0))) (char->integer #\0)))))))))
+			      (set! x (- (+ (* 10 x) (char->integer (string-ref c 0))) (char->integer #\0)))))))))
 	    
 	    (define (cursor-bounds)
 	      (let ((coords (cursor-coords)))
@@ -415,8 +389,8 @@
 		  (set! last-col (car bounds))
 		  (set! last-row (cdr bounds)))))
 	    
-	    ;; to enable mouse click coords (in xterm anyway), (format *stderr* "~C[?9h" #\escape)
-	    ;;    disable: (format *stderr* "~C[?9l" #\escape)
+	    ;; to enable mouse click coords (in xterm anyway), (fformat *stderr* "~C[?9h" #\escape)
+	    ;;    disable: (fformat *stderr* "~C[?9l" #\escape)
 	    ;;    this is X10_MOUSE -- probably better to use SET_ANY_EVENT_MOUSE 1003
 	    ;; while enabled, mouse selection instead sends coords to repl (so it's annoying)
 	    ;;    also it's sticky!  exit repl does not clear this flag so mouse is effectively dead
@@ -427,7 +401,7 @@
 	    
 	    ;; -------- display --------
 	    (define (display-prompt)
-	      (format *stderr* "~A" prompt-string))
+	      (fformat *stderr* "~A" prompt-string))
 	    
 	    (define (new-prompt)
 	      (set! cur-line "")
@@ -450,18 +424,19 @@
 		  (if (and (integer? red-par-pos)
 			   (<= start red-par-pos)
 			   (< red-par-pos end))
-		      (string-append
-		       (format #f (if (zero? start)
-				      (values "~A" prompt-string)
-				      (values "~NC" prompt-length #\space)))
-		       (format #f (if (= start red-par-pos)
-				      (values "~A~A"
-					      (bold (red "(")) 
-					      (substring cur-line (+ start 1) end))
-				      (values "~A~A~A"
-					      (substring cur-line start red-par-pos) 
-					      (bold (red "(")) 
-					      (substring cur-line (+ red-par-pos 1) end)))))
+		      (let-temporarily (((*s7* 'print-length) 10000))
+			(string-append
+			 (format #f (if (zero? start)
+					(values "~A" prompt-string)
+					(values "~NC" prompt-length #\space)))
+			 (format #f (if (= start red-par-pos)
+					(values "~A~A"
+						(bold (red "(")) 
+						(substring cur-line (+ start 1) end))
+					(values "~A~A~A"
+						(substring cur-line start red-par-pos) 
+						(bold (red "(")) 
+						(substring cur-line (+ red-par-pos 1) end))))))
 		      (format #f (if (zero? start)
 				     (values "~A~A" prompt-string (substring cur-line 0 end))
 				     (values "~NC~A" prompt-length #\space (substring cur-line start end))))))))
@@ -480,7 +455,7 @@
 	    
 	    (define (display-lines)
 	      (move-cursor prompt-row 0)
-	      (format *stderr* "~C[J" #\escape)
+	      (fformat *stderr* "~C[J" #\escape)
 	      (let ((len (length cur-line))
 		    (new-line ""))
 		(do ((line-end 0)
@@ -488,66 +463,9 @@
 		    ((> i len))
 		  (set! line-end (end-of-line i))
 		  (set! new-line (string-append new-line (display-line i (min (+ line-end 2) len)))))
-		(format *stderr* "~A" new-line)
+		(fformat *stderr* "~A" new-line)
 		(display-cursor)))
 	    
-	    
-	    ;; -------- help/debugging --------
-	    (define (one-line text)
-	      (if (not (string? text))
-		  text
-		  (let ((ntext (copy text)))
-		    (do ((i 0 (+ i 1)))
-			((= i (length ntext))
-			 ntext)
-		      (if (char=? (ntext i) #\newline) (set! (ntext i) #\|))))))
-	    
-	    (define (help c)
-	      (when (pair? (*repl* 'helpers))
-		(let ((coords (cursor-coords))
-		      (col (floor (/ last-col 2))))
-		  (move-cursor 1 col)
-		  (format *stderr* "+~NC" (- col 2) #\-)
-		  
-		  (do ((i 2 (+ i 1))                      ; put box in top right corner so we don't get trailing output as we scroll
-		       (lst (*repl* 'helpers) (cdr lst)))
-		      ((null? lst))
-		    (let ((str ((car lst) c)))
-		      (move-cursor i col)
-		      (format *stderr* "~C[K| ~A" #\escape (if (> (length str) col) (substring str 0 (- col 1)) str))))
-		  
-		  (move-cursor (+ 2 (length (*repl* 'helpers))) col)
-		  (format *stderr* "+~NC" (- col 2) #\-)
-		  (move-cursor (cdr coords) (car coords)))))
-	    
-	    (define (debug-help)
-	      (set! (*repl* 'helpers)
-		    (list
-		     (lambda (c) 
-		       (let ((cur-char (if (zero? (length cur-line))
-					   #\space
-					   (let ((c (cur-line (max 0 (min cursor-pos (- (length cur-line) 1))))))
-					     (if (char=? c #\newline) #\| c)))))
-			 (format #f "cursor: ~A, ~C, line: ~S"
-				 cursor-pos cur-char
-				 (one-line cur-line))))
-		     (lambda (c)
-		       (format #f "len: ~D, selection: ~S, previous: ~S" 
-			       (length cur-line) 
-			       (one-line selection) 
-			       (if (pair? prev-line) (one-line (cdar prev-line)) ())))
-		     (lambda (c)
-		       (format #f "cur-row: ~A, prompt-row: ~A" 
-			       cur-row 
-			       prompt-row))
-		     (lambda (c)
-		       (format #f "c: ~S ~D, start: ~A, end: ~A" 
-			       c
-			       (char->integer c)	
-			       (start-of-line cursor-pos)	
-			       (end-of-line cursor-pos))))))
-	    
-	    
 	    ;; -------- keymap(s) --------
 	    (define meta-keymap-functions (make-vector 256))
 	    (define keymap-functions (make-vector 256 #f))
@@ -619,7 +537,7 @@
 	      (set! cur-line (string-append cur-line (string #\space #\newline)))
 	      (set! cursor-pos (length cur-line))
 	      (when (= last-row (+ prompt-row cur-row))
-		(format *stderr* "~%")
+		(fformat *stderr* "~%")
 		(set! prompt-row (- prompt-row 1)))
 	      (set! cur-row (+ cur-row 1)))
 	    
@@ -715,7 +633,7 @@
 	    
 	    (set! (keymap-functions C-l) 
 		  (lambda (c)
-		    (format *stderr* "~C[H~C[J" #\escape #\escape)
+		    (fformat *stderr* "~C[H~C[J" #\escape #\escape)
 		    (new-prompt)))
 	    
 	    ;; -------- deletion
@@ -885,54 +803,59 @@
 			     
 			     ;; we want to add cur-line (copied) to the history buffer
 			     ;;   unless it is an on-going edit (missing close paren)
-			     (catch 'string-read-error
+			     (catch 'string-read-error ; this matches (throw #t 5) -- is this correct?
 			       
 			       (lambda ()
 				 (set! cursor-pos len)
 				 (if (or (= chars 1)
 					 (not (= input-fd terminal-fd)))
 				     (display-lines))
-				 (set! (history) cur-line)
+				 (set-history! cur-line)
 				 (set! m-p-pos 0)
 				 
-				 (with-repl-let
-				  (lambda ()
-				    ;; get the newline out if the expression does not involve a read error
-				    (let ((form (with-input-from-string cur-line #_read)))    ; not libc's read
-				      (newline *stderr*)
-				      (let ((val ((lambda args
-						    (if (pair? (cdr args))
-							(cons 'values args)
-							(car args)))
-						  (eval form (*repl* 'top-level-let)))))
-					(eval `(define ,(string->symbol (format #f "<~D>" (+ (length histtop) 1))) ',val) (rootlet))
-					(if unbound-case
-					    (set! unbound-case #f)
-					    (format *stderr* "~S~%" val)))))))
+				 ;; get the newline out if the expression does not involve a read error
+				 (let-temporarily (((hook-functions *missing-close-paren-hook*) (list badexpr)))
+				   (let ((form (with-input-from-string cur-line #_read)))    ; not libc's read
+				     (newline *stderr*)
+				     (let ((val (list (new-eval form (*repl* 'top-level-let))))) ; list, not lambda -- confuses trace!
+				       (if (or (null? val)   ; try to trap (values) -> #<unspecified>
+					       (and (unspecified? (car val))
+						    (null? (cdr val))))
+					   (set! val #<unspecified>)
+					   (set! val (if (pair? (cdr val))
+							 (cons 'values val)
+							 (car val))))
+						 
+				       (if unbound-case
+					   (set! unbound-case #f)
+					   (fformat *stderr* "~S~%" val))
+				       (eval `(define ,(string->symbol (format #f "<~D>" (+ (length histtop) 1))) ',val) (rootlet))))))
 			       
 			       (lambda (type info)
-				 (pop-history)               ; remove last history entry
-				 (append-newline)
-				 (return))))
+				 (if (eq? type 'string-read-error)
+				     (begin
+				       (pop-history)               ; remove last history entry
+				       (append-newline)
+				       (return))
+				     (apply throw type info)))))   ; re-raise error
 			   
 			   (lambda (type info)
 			     (with-let (unlet)
-			       (format *stderr* "~A:" (red "error"))
+			       (fformat *stderr* "~A:" (red "error"))
 			       (let ((op (*s7* 'print-length)))
 				 (if (< op 32) (set! (*s7* 'print-length) 32))
 				 (if (and (pair? info)
 					  (string? (car info)))
-				     (format *stderr* " ~A" (apply format #f info))
+				     (fformat *stderr* " ~A" (apply format #f info))
 				     (if (not (null? info))
-					 (format *stderr* " ~A" info)))
+					 (fformat *stderr* " ~A" info)))
 				 (if (< op 32) (set! (*s7* 'print-length) op)))
-			       (with-repl-let
-				(lambda ()
-				  (eval `(define ,(string->symbol (format #f "<~D>" (+ (length histtop) 1))) 'error) (rootlet))))
+			       (eval `(define ,(string->symbol (format #f "<~D>" (+ (length histtop) 1))) 'error) (rootlet))
 			       (newline *stderr*))))
 			 
 			 (push-line (copy cur-line))
 			 (new-prompt))))))
+
 	    
 	    ;; -------- escaped (Meta/Alt/arrow) keys
 	    (set! (keymap-functions Escape) 
@@ -965,7 +888,7 @@
 		    (set! cursor-pos (length cur-line))
 		    (let ((newlines (count-newlines cur-line)))
 		      (when (< last-row (+ prompt-row newlines))
-			(format *stderr* "~NC" (- (+ prompt-row newlines) last-row) #\newline)
+			(fformat *stderr* "~NC" (- (+ prompt-row newlines) last-row) #\newline)
 			(set! prompt-row (- prompt-row newlines)))
 		      (set! cur-row newlines)))))
 
@@ -986,7 +909,7 @@
 		(let ((old-index m-p-pos))
 		  (when (and (zero? m-p-pos) 
 			     (> (length cur-line) 0))
-		    (set! (history) cur-line)
+		    (set-history! cur-line)
 		    (set! m-p-pos -1))
 		  (set! m-p-pos (max (- m-p-pos 1) (- histsize)))
 		  (if (positive? (length (history m-p-pos)))
@@ -1051,13 +974,13 @@
 	    
 	    ;; -------- emacs --------
 	    (define (emacs-repl)
-	      ;; TODO: use the emacs language server protocol? (not our own rpc stuff or epc), for json, see json.scm 
-	      ;;   probably will need an argument/function? for repl to open the channel or whatever
+	      ;; someday maybe use the language server protocol? (not our own rpc stuff or epc), for json, see json.scm 
+	      ;;   probably will need an argument/function for repl to open the channel or whatever
 	      ;; also this does not resend the entire expression after editing
 	      ;;   and does not notice in-place edits
 	      ;;   can <cr> get entire expr?
 	      (let ((buf (c-pointer->string (calloc 512 1) 512)))
-		(format *stderr* "> ")
+		(fformat *stderr* "> ")
 		(do ((b (fgets buf 512 stdin) (fgets buf 512 stdin)))
 		    ((zero? (length b))
 		     (#_exit))
@@ -1068,24 +991,22 @@
 			       (= i len))
 			   (when (< i len)
 			     (let ((str (substring buf 0 (- (strlen buf) 1))))
-			       ;(format *stderr* "str: ~S~%" str)
+			       ;(fformat *stderr* "str: ~S~%" str)
 			       (catch #t
 				 (lambda ()
 				   (do ()
 				       ((= (string-length str) 0))
 				     (catch 'string-read-error
 				       (lambda ()
-					 (with-repl-let
-					  (lambda ()
-					    (format *stderr* "~S~%> " (eval-string str (*repl* 'top-level-let)))
-					    (set! str ""))))
+					 (fformat *stderr* "~S~%> " (eval-string str (*repl* 'top-level-let)))
+					 (set! str ""))
 				       (lambda (type info)
 					 (fgets buf 512 stdin)
 					 (set! str (string-append str " " (substring buf 0 (- (strlen buf) 1))))))))
 				 (lambda (type info)
 				   (set! str "")
 				   (apply format *stderr* info)
-				   (format *stderr* "~%> "))))))))))))
+				   (fformat *stderr* "~%> "))))))))))))
 
 	    ;; -------- rxvt et al --------
 	    (define (terminal-repl file)		
@@ -1099,9 +1020,10 @@
 		
 		(varlet (*repl* 'top-level-let) 
 		  :exit (let ((+documentation+ "(exit) resets the repl tty and exits the repl"))
-			  (lambda ()
-			    (newline *stderr*)
-			    (tty-reset))))
+			  (let-temporarily (((*s7* 'debug) 0))
+			    (lambda ()
+			      (newline *stderr*)
+			      (tty-reset)))))
 		
 		;; a "normal" terminal -- hopefully it accepts vt100 codes
 		(let ((buf (termios.make)))
@@ -1110,6 +1032,7 @@
 			  (let* ((c (make-string read-size #\null)) 
 				 (cc (string->c-pointer c))
 				 (ctr 0))
+			    (let-temporarily (((*s7* 'debug) 0))
 			    (lambda ()
 			      (call-with-exit
 			       (lambda (return)
@@ -1203,9 +1126,9 @@
 					 (lambda (type info)
 					   (set! chars 0)
 					   ;(move-cursor prompt-row prompt-col)
-					   (format *stderr* "internal error: ")
+					   (fformat *stderr* "internal error: ")
 					   (apply format *stderr* info)
-					   (format *stderr* "~%line ~A: ~A~%" ((owlet) 'error-line) ((owlet) 'error-code))
+					   (fformat *stderr* "~%line ~A: ~A~%" ((owlet) 'error-line) ((owlet) 'error-code))
 					   (set! chars 0)
 					   (set! ctr 0)
 					   (new-prompt)
@@ -1215,7 +1138,7 @@
 				 (let ((result (c ctr)))
 				   (set! ctr (+ ctr 1))
 				   result)))))))
-		  
+
 		  (set! input-fd (if (not file) 
 				     terminal-fd
 				     (open file O_RDONLY 0)))
@@ -1252,24 +1175,34 @@
 		      
 		      (lambda (type info)
 			;(move-cursor prompt-row prompt-col)
-			(format *stderr* "internal error: ")
+			(fformat *stderr* "internal error: ")
 			(apply format *stderr* info)
-			(format *stderr* "~%line ~A: ~A~%" ((owlet) 'error-line) ((owlet) 'error-code))
+			(fformat *stderr* "~%line ~A: ~A~%" ((owlet) 'error-line) ((owlet) 'error-code))
 			(set! chars 0)
-			(new-prompt)))))))
-	    
+			(new-prompt))))))))
+
 	    (define* (run file)
-	      ;; check for dumb terminal
-	      (if (or (zero? (isatty terminal-fd))        ; not a terminal -- input from pipe probably
-		      (string=? (getenv "TERM") "dumb"))  ; no vt100 codes -- emacs subjob for example
-		  (emacs-repl)                            ; TODO: restore support for the pipe case
-		  (terminal-repl file)))
+	      ;(when (*s7* 'history-enabled) (fill! (*s7* 'history) ()))
+	      (let-temporarily (((*s7* 'history-enabled) #f))
+		;; check for dumb terminal
+		(if (or (zero? (isatty terminal-fd))        ; not a terminal -- input from pipe probably
+			(string=? (getenv "TERM") "dumb"))  ; no vt100 codes -- emacs subjob for example
+		    (emacs-repl)                            ; TODO: restore support for the pipe case
+		    (terminal-repl file))))
 
 	    (curlet))))))
       
-      (define (save-repl) 
-	(call-with-output-file "save.repl" 
+      (define* (save-repl (file "save.repl"))
+	;; ((*repl* 'save-repl))
+	(call-with-output-file file
 	  (lambda (p) 
+	    (format p ";;; save-repl ~A~%~%" 
+		    (with-let (sublet *libc*)
+		      (let ((timestr (make-string 128))) 
+			(let ((len (strftime timestr 128 "%a %d-%b-%Y %H:%M:%S %Z"
+					     (localtime 
+					      (time.make (time (c-pointer 0 'time_t*)))))))
+			  (substring timestr 0 len)))))
 	    (format p "(for-each~%~NC~
                          (lambda (f)~%~NC~
                            (if (not (provided? f))~%~NC~
@@ -1280,10 +1213,14 @@
 	    (format p "(with-let (*repl* 'repl-let)~%")
 	    (((*repl* 'repl-let) 'write-history) p 2)
 	    (format p ")~%~%")
-	    (format p "~W" (*repl* 'top-level-let)))))
+	    (for-each (lambda (var)
+			(unless (eq? (car var) 'exit)
+			  (format p "(define ~S ~W)~%" (car var) (cdr var))))
+		      (*repl* 'top-level-let)))))
       
-      (define (restore-repl) 
-	(set! (*repl* 'top-level-let) (load "save.repl")))  
+      (define* (restore-repl (file "save.repl"))
+	(set! (*repl* 'top-level-let) (sublet (rootlet)))
+	(load file))
       ;; I think this could be a merge rather than a reset by using (with-let top-level-let (load ...))
       
       (set! keymap (repl-let 'keymap))
@@ -1297,14 +1234,62 @@
       (curlet))))
 
 ;; ((*repl* 'run))
+(set! (*s7* 'debug) old-debug)
 
 
 ;;; --------------------------------------------------------------------------------
 
+(define drop-into-repl
+  (let ((depth 0))
+    (lambda (call e)
+      (let ((C-q (integer->char 17)))      ; C-q to exit repl
+	(let-temporarily ((((*repl* 'keymap) C-q) (let-temporarily (((*s7* 'debug) 0)) 
+						    (lambda (c)
+						      (set! depth (- depth 1))
+						      (set! ((*repl* 'repl-let) 'all-done) #t))))
+			  ((*repl* 'top-level-let) e)
+			  ((*repl* 'prompt) (let-temporarily (((*s7* 'debug) 0)) 
+					      (lambda (num) 
+						(with-let (sublet (*repl* 'repl-let) :depth depth)
+						  (set! prompt-string (format #f "break~NC " depth #\>))
+						  (set! prompt-length (length prompt-string)))))))
+	  (set! depth (+ depth 1))
+	  (with-let (sublet (*repl* 'repl-let) :depth depth)
+	    (set! cur-line "")
+	    (set! red-par-pos #f)
+	    (set! cursor-pos 0)
+	    (set! prompt-string (format #f "break~NC " depth #\>))
+	    (set! prompt-length (length prompt-string)))
+	  (fformat *stderr* "break: ~A, C-q to exit break~%" call)
+	  ((*repl* 'run)))))))
+
+(define (debug.scm-init)
+  (set! ((funclet trace-in) '*debug-repl*) drop-into-repl))
+
+#|
+(define (open-file-with-catch filename)
+  (catch 'io-error ; is this correct?
+    (lambda ()
+      (open-input-file filename))
+    (lambda (type info)
+      (let ((new-filename #f)
+	    (err ((*libc* 'errno)))) ; get libc error indication (might be disk full or no write permission etc)
+	(drop-into-repl (format #f "can't open ~S, ~S" filename ((*libc* 'strerror) err)) (curlet))
+	(and new-filename
+	     (open-file-with-catch new-filename))))))
+|#
+
 (autoload 'lint "lint.scm")
 (autoload 'pretty-print "write.scm")
 (autoload '*libm* "libm.scm")
 (autoload '*libgsl* "libgsl.scm")
+(autoload 'trace "debug.scm")
+(autoload 'untrace "debug.scm")
+(autoload 'break "debug.scm")
+(autoload 'unbreak "debug.scm")
+(autoload 'watch "debug.scm")
+(autoload 'unwatch "debug.scm")
+(autoload 'ow! "stuff.scm")
 
 #|
 (define pwd
@@ -1322,7 +1307,7 @@
 		(let ((timestr (make-string 128))) 
 		  (let ((len (strftime timestr 128 "%a %d-%b-%Y %H:%M:%S %Z"
 				       (localtime 
-					(time.make (time (c-pointer 0)))))))
+					(time.make (time (c-pointer 0 'time_t*)))))))
 		    (substring timestr 0 len)))))))
     (openlet (inlet 'object->string pd 'let-ref-fallback pd))))
 |#
@@ -1352,8 +1337,6 @@
       (inlet 'object->string (lambda args
 			       (system ((*repl* 'repl-let) 'cur-line) #t))))))
 ;; (make-command ls)
-|#
-
 
 (define-macro (time expr)
   (let ((start (gensym)))
@@ -1362,9 +1345,14 @@
        (let ((end ((*libc* 'gettimeofday))))
 	 (+ (- (car end) (car ,start)) ; seconds
 	    (* 0.000001 (- (cadr end) (cadr ,start))))))))
+|#
 
+(define-macro (time expr) 
+  `(let ((start (*s7* 'cpu-time)))
+     ,expr 
+     (- (*s7* 'cpu-time) start)))
 
-(define apropos 
+(define apropos ; this misses syntax names (they aren't in rootlet)
   (let ((levenshtein 
 	 (lambda (s1 s2)
 	   (let ((L1 (length s1))
@@ -1410,7 +1398,8 @@
     
     (lambda* (name (e (*repl* 'top-level-let)))
       (let ((ap-name (if (string? name) name 
-			 (if (symbol? name) (symbol->string name)
+			 (if (symbol? name) 
+			     (symbol->string name)
 			     (error 'wrong-type-arg "apropos argument 1 should be a string or a symbol"))))
 	    (ap-env (if (let? e) e 
 			(error 'wrong-type-arg "apropos argument 2 should be an environment"))))
@@ -1432,7 +1421,7 @@
 	      'no-match
 	      (begin
 		(for-each (lambda (b)
-			    (format *stderr*
+			    (fformat *stderr*
 				    (if (zero? (cdr b)) "~C[1m~A~C[0m: ~S~%"                           ; black if exact match somewhere
 					(if (or (< (cdr b) 2) (not have-orange)) "~C[31m~A~C[0m: ~S~%" ; red for near miss
 					    "~C[38;5;208m~A~C[0m: ~S~%"))                              ; orange for less likely choices
@@ -1452,10 +1441,9 @@
 ;;; --------------------------------------------------------------------------------
 ;;; just a first stab at this -- where to put this code?
 ;;;
-;;; debug-help can be set running via: (((*repl* 'repl-let) 'debug-help))
 
 (define-expansion (repl-debug)
-  `(with-let (inlet :orig (curlet) :line ,(port-line-number) :func __func__)
+  `(with-let (inlet :orig (curlet) :line ,(port-line-number) :func (*function*))
      (format () "line ~D: ~A~{~^ ~A~}~%" 
 	     line (if (pair? func) (car func) func)
 	     (reverse
@@ -1589,7 +1577,7 @@ to post a help string (kinda tedious, but the helper list is aimed more at posti
 	(with-let (sublet (*repl* 'repl-let))
 	  (let ((next (next-char)))
 	    (when (char=? next (integer->char 6)) ; C-f
-	      (format *stderr* "~%load: ")
+	      (fformat *stderr* "~%load: ")
 	      ;; now recursive call: prompt="load: "
 	      (let ((old-prompt (*repl* 'prompt))
 		    (old-cur-line cur-line)
